/* 
   31337 recoder;)

   Copyright (C) 2002 Marek "Tilk" Materzok <tilk@lucifer.geos.pl>

   Ideas: Andrzej Lindnal <sznik@osama.geos.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "recoder.h"
#include "recoder_b64.h"
#include "recoder_numeric.h"
#include "recoder_roman.h"
#include "recoder_rot.h"
#include "recoder_morse.h"

#include <unistd.h>
#include <stdlib.h>
#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#endif
#include <string>
#include <sstream>
#include <fstream>
#include <iostream>
#include <list>
#include <vector>
#include <iomanip>
#include <algorithm>
#include <set>
#include <memory>
#ifndef _WIN32
#include <sys/wait.h>
#endif
#ifdef _WIN32
extern "C" 
{
int getopt(int argc, char * const argv[],const char *optstring);
extern char *optarg;
extern int optind, opterr, optopt;
};
#include <windows.h>
#endif
#include "config.h"


using namespace std;

list<Recoder *> recodes;
bool encode = true;
ostream *ostr = &cout;
istream *istr = &cin;

Input::Input()
{
	operator=(&cin);
}

void
Input::get(char &zn)
{
	if (bpos)
	{
		zn = buffer[8192 - bpos - prz];
		--bpos;
		return;
	}
	if (!*str)
	{
		bpos = -1;
		return;
	}
	str->read(buffer + 4096, 4096);
	bpos = str->gcount();
	prz = 4096 - bpos;
	zn = buffer[4096];
	bpos--;
}

void
Input::putback(char zn)
{
	buffer[8192 - (++bpos) - prz] = zn;
}

void
Input::putback(const std::string &str)
{
	std::copy(str.begin(), str.end(), buffer + 8192 - bpos - prz - str.size());
	bpos += str.size();
}

void
Input::operator=(istream *is)
{
	str = is;
	bpos = 0;
	prz = 0;
	completed = false;
}

Input::operator bool() const
{
	return *str || bpos != -1;
}

istream *
Input::getstr()
{
	return str;
}

void 
Options::Parse(const string &str)
{
	for (string::const_iterator iter = str.begin(); iter != str.end(); )
	{
		string::const_iterator eiter = ::find(iter, str.end(), ',');
		string subs(iter, eiter);
		AddOpt(subs);
		iter = eiter;
		if (iter != str.end()) ++iter;
	}
}

void 
Options::AddOpt(const string &str)
{
	insert(str);
}

void 
Options::DelOpt(const string &str)
{
	erase(str);
}
		
bool 
Options::HasOpt(const string &str)
{
	return count(str) > 0;
}
	
string 
Options::OptVal(const string &p_str)
{
	string str(p_str);
	str += "=";
	set<string>::iterator iter = lower_bound(str);
	if (iter != end() && iter->substr(0, str.size()) == str)
		return iter->substr(str.size(), string::npos);
	return string();
}

int 
Options::IntOptVal(const string &p_str)
{
	stringstream sstr;
	int ret;
	string str = OptVal(p_str);
	sstr << str;
	sstr >> ret;
	return ret;
}

RecoderError::RecoderError(const string &str) throw()
{
	descr = "Recoder error: " + str;
}

RecoderError::~RecoderError() throw()
{
}

const char *
RecoderError::what() const throw()
{
	return descr.c_str();
}

Recoder::Recoder(): opt_scanned(false)
{
	istr = &cin;
	ostr = &cout;
}

void
Recoder::ScanOpts()
{
};

void
Recoder::ParseOptions(const string &opts)
{ 
	options.Parse(opts);
	opt_scanned = false;
};

void
Recoder::Encode()
{
	if (!opt_scanned)
	{
		ScanOpts();
		opt_scanned = true;
	}
	DoEncode();
	ostr->flush();
}

void
Recoder::Decode()
{
	if (!opt_scanned)
	{
		ScanOpts();
		opt_scanned = true;
	}
	DoDecode();
	ostr->flush();
}

void
WorkRecode(Recoder *rc)
{
	if (encode) rc->Encode();
	else rc->Decode();
}

void
DoRecode()
{
	if (recodes.size() == 1)
	{
		recodes.front()->istr = istr;
		recodes.front()->ostr = ostr;

		WorkRecode(recodes.front());
		return;
	}

	if (!encode) reverse(recodes.begin(), recodes.end());

	recodes.front()->istr = istr;
	recodes.back()->ostr = ostr;

#ifndef _WIN32
	int pipes[2];

	pipe(pipes);
	if (fork() == 0)
	{
		close(1);
		dup2(pipes[1], 1);
		close(pipes[1]);
		WorkRecode(recodes.front());
		close(1);
		exit(0);
	}

	int readfd = pipes[0];

	for (list<Recoder *>::iterator iter = ++recodes.begin();
			iter != --recodes.end(); ++iter)
	{
		close(pipes[1]);
		pipe(pipes);
		if (fork() == 0)
		{
			close(0);
			close(1);
			dup2(readfd, 0);
			dup2(pipes[1], 1);
			close(readfd);
			close(pipes[1]);
			WorkRecode(*iter);
			close(1);
			exit(0);
		}
		close(readfd);
		readfd = pipes[0];
	}

	close(pipes[1]);

	if (fork() == 0)
	{
		close(0);
		dup2(readfd, 0);
		WorkRecode(recodes.back());
		exit(0);
	}

	close(pipes[0]);

	for (int n = 0; n < recodes.size(); n++) wait(0);
#else
	// TODO: threading implementation
	auto_ptr<stringstream> sstr_r, sstr_w, sstr_tmp;
	sstr_r.reset(new stringstream());
	sstr_w.reset(new stringstream());
	
	recodes.front()->ostr = sstr_w.get();
	WorkRecode(recodes.front());
	
	for (list<Recoder *>::iterator iter = ++recodes.begin();
			iter != --recodes.end(); ++iter)
	{
		sstr_tmp = sstr_r;
		sstr_r = sstr_w;
		sstr_w = sstr_tmp;
		(*iter)->istr = sstr_r.get();
		(*iter)->ostr = sstr_w.get();
		WorkRecode(*iter);
	}

	recodes.back()->istr = sstr_w.get();
	WorkRecode(recodes.back());

#endif

	for (list<Recoder *>::iterator iter = recodes.begin();
			iter != recodes.end(); ++iter)
	{
		delete *iter;
	}

	recodes.clear();
}

void
Usage(char *arg0)
{
	cerr << "31337 recoder ;)\n";
	cerr << _("usage:\n"); 
	cerr << "\t" << arg0 << _(" -e [args]\t= encode from stdin\n");
	cerr << "\t" << arg0 << _(" -ea [args] text\t= encode arguments\n");
	cerr << "\t" << arg0 << _(" -ef [args] files...\t= encode from files\n");
	cerr << "\t" << arg0 << _(" -d [args]\t= decode from stdin\n");
	cerr << "\t" << arg0 << _(" -da [args] text\t= decode arguments\n");
	cerr << "\t" << arg0 << _(" -df [args] files...\t= decode from files\n");
	cerr << _("\narguments:\n");
	cerr << _("\t-r num - recode with ROT-num (Caesar shift)\n");
	cerr << _("\t-n num - recode to numeric system num (2-36)\n");
	cerr << _("\t-B - recode with Base64\n");
	cerr << _("\t-H - recode to escaped hexadecimal\n");
	cerr << _("\t-O - recode to escaped octal\n");
	cerr << _("\t-R - recode to Roman numbers\n");
	cerr << _("\t-m - recode to Morse alphabet\n");
	cerr << _("\t-o file - output to file\n");
	cerr << _("\t-x options - pass options to codecs (look to manpage)\n");
	cerr << "\n";
}

void
Version()
{
	cerr << _("31337 recoder ;) version ") << VERSION "\n";
	cerr << "Copyleft 2002 Marek \"Tilk\" Materzok <tilk@lucifer.geos.pl>\n";
	cerr << _("Ideas: Andrzej Lindnal <sznik@osama.geos.pl>\n");
	cerr << _("Licensed under GNU GPL version 2\n");
}

int
main(int argc, char **argv)
{
	string ofile;
	bool f_stdin = true;
	bool f_file = false;

#ifdef HAVE_GETOPT_LONG
	static struct option long_options[] =
	{
		{"encode", false, 0, 'e'},
		{"decode", false, 0, 'd'},
		{"recode-arg", false, 0, 'a'},
		{"recode-file", false, 0, 'f'},
		{"output-file", true, 0, 'o'},
		{"numeric", true, 0, 'n'},
		{"rot", true, 0, 'r'},
		{"escaped-hex", false, 0, 'H'},
		{"escaped-octal", false, 0, 'O'},
		{"help", false, 0, 'h'},
		{"base64", false, 0, 'B'},
		{"version", false, 0, 'v'},
		{"roman", false, 0, 'R'},
		{"morse", false, 0, 'm'},
		{"options", true, 0, 'x'},
		0
	};
#endif
#ifdef HAVE_GETTEXT
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
#endif

	static const char *opt_list = "edEDn:hHOr:Bafo:vRx:m";
	int c;

	try
	{
#ifdef HAVE_GETOPT_LONG
	while ((c = getopt_long(argc, argv, opt_list, long_options, 0)) != -1)
#else
	while ((c = getopt(argc, argv, opt_list)) != -1)
#endif
	{
		switch(c)
		{
			case 'v':
				Version();
				return 0;
			case 'h':
				Usage(argv[0]);
				return 0;
			case 'e':
				encode = true;
				f_stdin = true;
				break;
			case 'd':
				encode = false;
				f_stdin = true;
				break;
			case 'a':
				f_stdin = false;
				break;
			case 'f':
				f_file = true;
				break;
			case 'n':
				recodes.push_back(new NumericRecoder());
				recodes.back()->ParseOptions(string("base=")+optarg);
				break;
			case 'r':
				recodes.push_back(new ROTRecoder());
				recodes.back()->ParseOptions(string("key=")+optarg);
				break;
			case 'o':
				ofile = optarg;
				break;
			case 'H':
				recodes.push_back(new NumericRecoder());
				recodes.back()->ParseOptions("base=16,startspace,space=\\x");
				break;
			case 'O':
				recodes.push_back(new NumericRecoder());
				recodes.back()->ParseOptions("base=8,startspace,space=\\0");
				break;
			case 'B':
				recodes.push_back(new B64Recoder());
				break;
			case 'R':
				recodes.push_back(new RomanRecoder());
				break;
			case 'x':
				if (recodes.empty())
				{
					cerr << _("No codec specified for -x.\n");
					return 1;
				}
				recodes.back()->ParseOptions(optarg);
				break;
			case 'm':
				recodes.push_back(new MorseRecoder());
				break;
			case '?':
				return 1;
		}
	}
	
	if (recodes.empty())
	{
		cerr << _("No codecs specified. Use ") << argv[0] << 
			_(" -h for help.\n");
		return 1;
	}

	if (ofile.size() > 0)
	{
		ostr = new ofstream(ofile.c_str());
		if (!*ostr)
		{
			cerr << _("Can't open file ") << ofile << _(" for writing!\n");
			return 1;
		}
	}
	
	if (!f_stdin && !f_file)
	{
		stringstream *sstr = new stringstream();
		for (int i = optind; i < argc; i++)
		{
			*sstr << argv[i];
			if (i != argc - 1) *sstr << " ";
		}
		istr = sstr;
		DoRecode();
		delete sstr;
	}
	else
	if (f_file)
	{
		for (int i = optind; i < argc; i++)
		{
			ifstream *ifstr = new ifstream(argv[i]);
			istr = ifstr;
			if (!istr)
			{
				cerr << _("Can't open file ") << argv[i] << "!\n";
				continue;
			}
			DoRecode();
			delete ifstr;
		}
	}
	else
	{
		DoRecode();
	}
	ostr->flush();
	cerr << "\n";
	}
	catch(std::exception &err)
	{
		cerr << err.what() << "\n";
		return 1;
	}
	
	return 0;
}

