/* 
   31337 recoder;)

   Copyright (C) 2002 Marek "Tilk" Materzok <tilk@lucifer.geos.pl>

   Ideas: Andrzej Lindnal <sznik@osama.geos.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

#ifndef __eleet_recoder_h
#define __eleet_recoder_h

#ifdef __GNUC__
#ifndef __USE_GNU
#define __USE_GNU
#endif
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#endif

#ifdef HAVE_GETTEXT
#include <libintl.h>
#define _(X) gettext(X)
#else
#define _(X) X
#endif

#include <set>
#include <string>
#include <iostream>

class Options: private std::set<std::string>
{
	public:
		void Parse(const std::string &str);
		void AddOpt(const std::string &str);
		void DelOpt(const std::string &str);
		bool HasOpt(const std::string &str);
		std::string OptVal(const std::string &str);
		int IntOptVal(const std::string &str);
};

class RecoderError: public std::exception
{
	public:
		RecoderError(const std::string &str) throw();
		~RecoderError() throw();

		const char *what() const throw();
		
	private:
		std::string descr;
};

class Input
{
	public:
		Input();
		void get(char &zn);
		void putback(char zn);
		void putback(const std::string &str);
		void operator=(std::istream *is);
		operator bool() const;
		std::istream *getstr();

	private:
		std::istream *str;
		char buffer[8192];
		int bpos;
		int prz;
		bool completed;
};

class Recoder
{
	public:
		Recoder();

		void ParseOptions(const std::string &opts);

		void Encode();
		void Decode();

		std::ostream *ostr;
		Input istr;
		
	protected:
		virtual void DoEncode() = 0;
		virtual void DoDecode() = 0;

		virtual void ScanOpts();
		
		Options options;

	private:
		bool opt_scanned;
};

#endif

