/* 
   31337 recoder;)

   Copyright (C) 2002 Marek "Tilk" Materzok <tilk@lucifer.geos.pl>

   Ideas: Andrzej Lindnal <sznik@osama.geos.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "recoder.h"
#include "recoder_morse.h"

#include <map>

namespace
{
	std::map<char, std::string> enc_map;
	std::map<std::string, char> dec_map;
};

MorseRecoder::MorseRecoder()
{
	pass_uncoded = false;
}

void
MorseRecoder::ScanOpts()
{
	pass_uncoded = options.HasOpt("passuc");
}

void
MorseRecoder::Init()
{
	const char  chars[] = {
		'A',
		'B',
		'C',
		'D',
		'E',
		'F',
		'G',
		'H',
		'I',
		'J',
		'K',
		'L',
		'M',
		'N',
		'O',
		'P',
		'Q',
		'R',
		'S',
		'T',
		'U',
		'V',
		'W',
		'X',
		'Y',
		'Z',
		'0',
		'1',
		'2',
		'3',
		'4',
		'5',
		'6',
		'7',
		'8',
		'9',
		'.',
		',',
		':',
		'?',
		'-',
		'"',
		0
	};
	const char *codes[] = {
		".-", 
		"-...", 
		"-.-.", 
		"-..", 
		".", 
		"..-.", 
		"--.", 
		"....", 
		"..", 
		".---", 
		"-.-", 
		".-..", 
		"--", 
		"-.", 
		"---", 
		".--.", 
		"--.-", 
		".-.", 
		"...", 
		"-", 
		"..-", 
		"...-", 
		".--", 
		"-..-", 
		"-.--", 
		"--..", 
		"-----",
		".----",
		"..---",
		"...--",
		"....-",
		".....",
		"-....",
		"--...",
		"---..",
		"----.",
		".-.-.-",
		"--..--",
		"---...",
		"..--..",
		"-....-",
		".-..-.",
		0
	};
	const char *chr;
	const char **code;
	for (chr = chars, code = codes;
			*code != 0; chr++, code++)
	{
		enc_map[*chr] = *code;
		dec_map[*code] = *chr;
	};
}

std::string
MorseRecoder::EncodeChar()
{
	std::string ret;
	char ch;
	istr.get(ch);
	if (!istr) return ret;
	std::map<char, std::string>::iterator iter = enc_map.find(::toupper(ch));
	if (iter != enc_map.end())
	{
		if (!omit_lspacer) ret = " ";
		ret += iter->second;
		omit_lspacer = false;
	} 
	else
	{
		if (ch == ' ')
		{
			ret = "|";
			omit_lspacer = true;
		}
		else
		if (pass_uncoded)
		{
			ret = ch;
		}
	}
	return ret;
}

void
MorseRecoder::DoEncode()
{
	if (enc_map.empty()) Init();
	omit_lspacer = true;

	while(istr)
	{
		std::string str = EncodeChar();
		if (!str.empty()) *ostr << str;
	}
}

char
MorseRecoder::DecodeChar()
{
	std::string mch;
	while (istr)
	{
		char ch;
		istr.get(ch);
		if (!istr) break;
		if (ch == ' ')
		{
			if (mch.empty()) continue;
			else break;
		}
		if (!(ch == '-' || ch == '.'))
		{
			if (mch.empty())
			{
				if (ch == '|') return ' ';
				return ch;
			}
			else
			{
				istr.putback(ch);
				break;
			}
		}
		mch.push_back(ch);
	}
	std::map<std::string, char>::iterator iter = dec_map.find(mch);
	if (iter == dec_map.end()) throw 1;
	return iter->second;
}

void
MorseRecoder::DoDecode()
{
	if (enc_map.empty()) Init();
	
	try
	{
		while (istr)
		{
			*ostr << DecodeChar();
		}
	}
	catch(int)
	{
	}
}

