/* 
   31337 recoder;)

   Copyright (C) 2002 Marek "Tilk" Materzok <tilk@lucifer.geos.pl>

   Ideas: Andrzej Lindnal <sznik@osama.geos.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "recoder.h"
#include "recoder_numeric.h"

NumericRecoder::NumericRecoder()
{
	numsys = 0;
}

char
NumericRecoder::ValToChar(int val)
{
	const char *chars = "0123456789abcdefghijklmnopqrstuvwxyz";
	if (val >= 36 || val < 0) return ' ';
	return chars[val];
};

int
NumericRecoder::CharToVal(char zn)
{
	int vals[] = {
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		0,	1,	2,	3,	4,	5,	6,	7,	8,	9,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	10,	11,	12,	13,	14,	15,	16,	17,	18,	19,	20,	21,	22,	23,	24,
		25,	26,	27,	28,	29,	30,	31,	32,	33,	34,	35,	-1,	-1,	-1,	-1,	-1,
		-1,	10,	11,	12,	13,	14,	15,	16,	17,	18,	19,	20,	21,	22,	23,	24,
		25,	26,	27,	28,	29,	30,	31,	32,	33,	34,	35,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,
		-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,	0};
	return vals[static_cast<char>(zn)];
}

void
NumericRecoder::ScanOpts()
{
	numsys = options.IntOptVal("base");
	if (numsys < 2 || numsys > 36)
		throw RecoderError(_("Supported numeric system bases: 2 - 36.\n"));
	nozeroes = options.HasOpt("nozeroes");	
	nospaces = options.HasOpt("nospaces");	
	startspace = options.HasOpt("startspace");	
	endspace = options.HasOpt("endspace");
	
	spacer = options.OptVal("space");
	if (spacer.empty()) spacer = " ";
	
	if (nospaces && nozeroes) nospaces = false;

	dpc = 0;
	int n = 1;
	while (n < 255)
	{
		n *= numsys;
		dpc++;
	}
	maxpow = n / numsys;
}

std::string
NumericRecoder::EncodeDigit()
{
	unsigned char zn;
	char uzn;
	istr.get(uzn);
	zn = uzn;
	if (!istr) return std::string();

	std::string ret;
	ret.resize(dpc);

	int xpow = maxpow;
	int firstch = 0;
	
	for (int n = 0; n < dpc; n++, xpow /= numsys)
	{
		int val = 0;
		while (zn >= xpow)
		{
			zn -= xpow;
			val++;
		}
		if (val == 0 && firstch == 0 && nozeroes) continue;
		if (firstch == 0 && nozeroes)
		{
			firstch = n;
			nozeroes = false;
		}
		ret[n - firstch] = ValToChar(val); 
	}

	ret.resize(dpc - firstch);
	return ret;
}

void
NumericRecoder::DoEncode()
{
	bool start = true;

	while (istr)
	{
		std::string str = EncodeDigit();
		if (str.size() == 0) break;
		if (!nospaces && (!start || startspace)) 
			*ostr << spacer;
		start = false;
		*ostr << str;
	}

	if (endspace) *ostr << spacer;
}

char
NumericRecoder::DecodeDigit()
{
	int dread;
	unsigned char res = 0;
	
	for (dread = 0; dread < dpc; dread++)
	{
		if (!istr) break;
		char zn;
		
		istr.get(zn);

		if (spacer.size() > 1 && zn == spacer[0])
		{
			std::string rsp;
			rsp.push_back(zn);
			int rtn = 1;
			while (rsp.size() < spacer.size() && istr)
			{
				istr.get(zn);
				rsp.push_back(zn);
				rtn++;
				if (zn != spacer[rtn - 1]) break;
			}
			if (rsp != spacer)
			{
				istr.putback(rsp);
			}
			istr.get(zn);
		}

		int cval = CharToVal(zn);
		if (cval == -1 || cval >= numsys || !istr)
		{
			if (dread == 0)
			{
				dread--;
				continue;
			}
			dread = dpc;
			break;
		}
		res *= numsys;
		res += CharToVal(zn);
	}
	if (dread != dpc) throw 1;
	return static_cast<char>(res);
}

void
NumericRecoder::DoDecode()
{
	try
	{
		while (istr)
		{
			*ostr << DecodeDigit();
		}
	}
	catch(int)
	{
	}
}

