/* 
   31337 recoder;)

   Copyright (C) 2002 Marek "Tilk" Materzok <tilk@lucifer.geos.pl>

   Ideas: Andrzej Lindnal <sznik@osama.geos.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "recoder.h"
#include "recoder_roman.h"

std::string
RomanRecoder::EncodeDigit()
{
	std::string ret;
	unsigned char zn;
	char uzn;
	istr.get(uzn);
	zn = uzn;
	if (!istr) return ret;
	if (zn == 0)
	{
		ret = "-";
		return ret;
	}

	char M[4] = {'I', 'X', 'C', 'M'};
	char D[4] = {'V', 'L', 'D', '-'};

	int coef, deg;

	for (coef = 100, deg = 3; deg > 0; deg--, coef /= 10)
	{
		while (zn > 10 * coef)
		{
			ret.push_back(M[deg]);
			zn -= 10 * coef;
		}

		if (zn >= 9 * coef)
		{
			ret.push_back(M[deg - 1]);
			ret.push_back(M[deg]);
			zn -= 9 * coef;
		}

		if (zn >= 5 * coef)
		{
			ret.push_back(D[deg - 1]);
			zn -= 5 * coef;
		}

		if (zn >= 4 * coef)
		{
			ret.push_back(M[deg - 1]);
			ret.push_back(D[deg - 1]);
			zn -= 4 * coef; 
		}
	}
	while (zn--) ret.push_back('I');

	return ret;
}

void
RomanRecoder::DoEncode()
{
	while(istr)
	{
		std::string str = EncodeDigit();
		if (!str.empty()) *ostr << str << " ";
	}
}

char
RomanRecoder::DecodeDigit()
{
	unsigned char val = 0;
	int lastd = 0;
	bool complete = false;
	bool loop = false;

	while (!complete)
	{
		int d = 0;
		char ch;
		istr.get(ch);
		if (!(istr || loop)) throw 1;
		if (!istr || isspace(ch))
		{
			ch = ' ';
			complete = true;
		}
		loop = true;
		ch = toupper(ch);
		switch(ch)
		{
			case 'I': d = 1; break;
			case 'V': d = 5; break;
			case 'X': d = 10; break;
			case 'L': d = 50; break;
			case 'C': d = 100; break;
			case 'D': d = 500; break;
			case 'M': d = 1000; break;
			case '-': d = 0; break;
			default: loop = false;
		}
		if (lastd < d) val -= lastd;
		else val += lastd;
		lastd = d;
	}
	
	return static_cast<char>(val);
}

void
RomanRecoder::DoDecode()
{
	try
	{
		while (istr)
		{
			*ostr << DecodeDigit();
		}
	}
	catch(int)
	{
	}
}

